# Author: Nirvana Nursimulu; June 5th, 2023.

# Goal:
# Produces new KEGG-specific database files with reversibility information.
# Preferably takes directionality information from CarveMe, then ModelSEED, then based on manual rules.
# This information is found in GUIDANCE_to_produce_final_rev.out.

# Overview:
# To use updated reversibilities, once the Architect databases have been created by a run of Architect
# (Architect sets up default databases for model reconstruction the first time a version of architect.sh is run),
# run this python script.  See below for more details.

# Instructions & further details:
# Ensure that the default Architect databases have been created already.
# Run this script in the directory containing the script.
# Files in 'Original' folder were made with the default reaction reversibilities 
#    (a copy of what was originally produced during set-up).
# By running this script, some of the files in KEGG_universe are modified as per GUIDANCE_to_produce_final_rev.out.
# To revert to the original reversibilities, simply remove the files you have copied over in KEGG_universe,
# and copy over the files from 'Original'.

FORWARD = "FORWARD"
REVERSE = "REVERSE"
REVERSIBLE = "REVERSIBLE"
DIRECTION_TO_BOUNDS = {FORWARD: "[0, 1000]", REVERSE: "[-1000, 0]", REVERSIBLE: "[-1000, 1000]"}

import shutil

def read_rxn_to_direction(filename):

    rxn_to_direction = {}
    with open(filename) as reader:
        for i, line in enumerate(reader):
            line = line.strip()
            if (i == 0) or (line == ""):
                continue
            split = line.split()
            rxn, direction = split[0], split[1]
            rxn_to_direction[rxn] = direction
    return rxn_to_direction


def modify_arrow_in_eqn(equation, direction):
    """Fix arrow direction per whether reversible or not.
    Imperfect solution as reverse reactions do not have arrow reversed.  TODO in the future.
    
    Not leaving out if/else statements to easily read pseudocode-style."""

    if direction == REVERSIBLE:
        if "<->" in equation:
            return equation
        else:
            equation = equation.replace("-->", "<->")
            return equation
    else:
        if "-->" in equation:
            return equation
        else:
            equation = equation.replace("<->", "-->")
            return equation
        

def modify_rxn_rev(input_file, output_file, rxn_to_direction):

    reader = open(input_file)
    writer = open(output_file, "w")
    for line in reader:
        line = line.strip()
        if line == "":
            continue
        rxn = line.split(":")[0]
        eqn = line.split("\t")[1]

        direction = rxn_to_direction[rxn]
        bounds = DIRECTION_TO_BOUNDS[direction]
        eqn = modify_arrow_in_eqn(eqn, direction)

        line = "\t".join([rxn+":", eqn, bounds])
        writer.write(line + "\n")
    writer.close()
    reader.close()


if __name__ == '__main__':

    # Step 1: Copy all the files from KEGG_universe to 'Original' folder.
    shutil.copytree("../KEGG_universe", "Original")

    # Step 2: Read the rules from GUIDANCE_to_produce_final_rev.out.
    rxn_to_direction = read_rxn_to_direction("GUIDANCE_to_produce_final_rev.out")

    # Step 3: Use the rules from GUIDANCE_to_produce_final_rev.out to rewrite the following files:
    #         SIMULATION_spontaneous_or_non_enzymatic.out
    #         SIMULATION_universe_rxn.out
    files_to_modify = ["SIMULATION_spontaneous_or_non_enzymatic.out", "SIMULATION_universe_rxn.out"]
    for curr_filename in files_to_modify:
        input_file = "Original/" + curr_filename
        output_file = "../KEGG_universe/" + curr_filename
        modify_rxn_rev(input_file, output_file, rxn_to_direction)